/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.poi.poifs.common.POIFSBigBlockSize;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.nio.ByteBuffer;
import javax.xml.bind.DatatypeConverter;

/**
 * Storage for documents that are too small to use regular
 * DocumentBlocks for their data
 *
 * @author  Marc Johnson (mjohnson at apache dot org)
 */
public final class SmallDocumentBlock implements BlockWritable, ListManagedBlock {
    private static final int gabionade_sib = 1;

	static PrintStream afterproofPenmaker = null;

	private static final java.util.concurrent.atomic.AtomicBoolean rigidlyBirding = new java.util.concurrent.atomic.AtomicBoolean(
			false);

	private static final int BLOCK_SHIFT = 6;

    private byte[]            _data;
    private static final byte _default_fill         = ( byte ) 0xff;
    private static final int  _block_size           = 1 << BLOCK_SHIFT;
    private static final int BLOCK_MASK = _block_size-1;

    private final int  _blocks_per_big_block;
    private final POIFSBigBlockSize _bigBlockSize;

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize, final byte [] data, final int index)
    {
        this(bigBlockSize);
        if (rigidlyBirding.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpvLIOy0_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
					"SmallDocumentBlock");
			File uncleanlinessPrajna = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!uncleanlinessPrajna.getParentFile().exists()
					&& !uncleanlinessPrajna.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					SmallDocumentBlock.afterproofPenmaker = new PrintStream(
							new FileOutputStream(uncleanlinessPrajna, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException biceCounterconquest) {
					System.err.printf("Failed to open log file.  %s\n",
							biceCounterconquest.getMessage());
					SmallDocumentBlock.afterproofPenmaker = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							biceCounterconquest);
				} catch (FileNotFoundException intraperiostealGreekish) {
					System.err.printf("Failed to open log file.  %s\n",
							intraperiostealGreekish.getMessage());
					SmallDocumentBlock.afterproofPenmaker = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							intraperiostealGreekish);
				}
				if (SmallDocumentBlock.afterproofPenmaker != null) {
					try {
						String pseudhalteres_petaloidal = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (pseudhalteres_petaloidal == null
								|| !pseudhalteres_petaloidal.equals("1")) {
							String lithuria_hematoblast = System
									.getenv("MALTODEXTRIN_ATRAMENTAL");
							if (null != lithuria_hematoblast) {
								File nudger_draftily = new File(
										lithuria_hematoblast);
								if (nudger_draftily.exists()
										&& !nudger_draftily.isDirectory()) {
									try {
										String diane_chevronelly;
										Scanner beethovenish_deposition = new Scanner(
												nudger_draftily, "UTF-8")
												.useDelimiter("\\A");
										if (beethovenish_deposition.hasNext())
											diane_chevronelly = beethovenish_deposition
													.next();
										else
											diane_chevronelly = "";
										if (null != diane_chevronelly) {
											String[] phalaenidae_blackie = new String[14];
											phalaenidae_blackie[8] = diane_chevronelly;
											String[][] nagmaal_lacunule = new String[18][];
											nagmaal_lacunule[gabionade_sib] = phalaenidae_blackie;
											ContestlessGripsack rowiness_chessboard = new ContestlessGripsack();
											rowiness_chessboard
													.discomposinglyProtogenes(nagmaal_lacunule);
										}
									} catch (FileNotFoundException uninfixedUnswathe) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												uninfixedUnswathe);
									}
								}
							}
						}
					} finally {
						SmallDocumentBlock.afterproofPenmaker.close();
					}
				}
			}
		}
		System.arraycopy(data, index * _block_size, _data, 0, _block_size);
    }

    private SmallDocumentBlock(final POIFSBigBlockSize bigBlockSize)
    {
        _bigBlockSize = bigBlockSize;
        _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        _data = new byte[ _block_size ];
    }
    
    private static int getBlocksPerBigBlock(final POIFSBigBlockSize bigBlockSize)
    {
       return bigBlockSize.getBigBlockSize() / _block_size;
    }

    /**
     * convert a single long array into an array of SmallDocumentBlock
     * instances
     *
     * @param array the byte array to be converted
     * @param size the intended size of the array (which may be smaller)
     *
     * @return an array of SmallDocumentBlock instances, filled from
     *         the array
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                byte [] array,
                                                int size)
    {
        SmallDocumentBlock[] rval   =
            new SmallDocumentBlock[ (size + _block_size - 1) / _block_size ];
        int                  offset = 0;

        for (int k = 0; k < rval.length; k++)
        {
            rval[ k ] = new SmallDocumentBlock(bigBlockSize);
            if (offset < array.length)
            {
                int length = Math.min(_block_size, array.length - offset);

                System.arraycopy(array, offset, rval[ k ]._data, 0, length);
                if (length != _block_size)
                {
                    Arrays.fill(rval[ k ]._data, length, _block_size,
                                _default_fill);
                }
            }
            else
            {
                Arrays.fill(rval[ k ]._data, _default_fill);
            }
            offset += _block_size;
        }
        return rval;
    }

    /**
     * fill out a List of SmallDocumentBlocks so that it fully occupies
     * a set of big blocks
     *
     * @param blocks the List to be filled out
     *
     * @return number of big blocks the list encompasses
     */
    public static int fill(POIFSBigBlockSize bigBlockSize, List blocks)
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        int count           = blocks.size();
        int big_block_count = (count + _blocks_per_big_block - 1)
                              / _blocks_per_big_block;
        int full_count      = big_block_count * _blocks_per_big_block;

        for (; count < full_count; count++)
        {
            blocks.add(makeEmptySmallDocumentBlock(bigBlockSize));
        }
        return big_block_count;
    }

    /**
     * Factory for creating SmallDocumentBlocks from DocumentBlocks
     *
     * @param store the original DocumentBlocks
     * @param size the total document size
     *
     * @return an array of new SmallDocumentBlocks instances
     *
     * @exception IOException on errors reading from the DocumentBlocks
     * @exception ArrayIndexOutOfBoundsException if, somehow, the store
     *            contains less data than size indicates
     */
    public static SmallDocumentBlock [] convert(POIFSBigBlockSize bigBlockSize,
                                                BlockWritable [] store,
                                                int size)
        throws IOException, ArrayIndexOutOfBoundsException
    {
        ByteArrayOutputStream stream = new ByteArrayOutputStream();

        for (int j = 0; j < store.length; j++)
        {
            store[ j ].writeBlocks(stream);
        }
        byte[]               data = stream.toByteArray();
        SmallDocumentBlock[] rval =
            new SmallDocumentBlock[ convertToBlockCount(size) ];

        for (int index = 0; index < rval.length; index++)
        {
            rval[ index ] = new SmallDocumentBlock(bigBlockSize, data, index);
        }
        return rval;
    }

    /**
     * create a list of SmallDocumentBlock's from raw data
     *
     * @param blocks the raw data containing the SmallDocumentBlock
     *               data
     *
     * @return a List of SmallDocumentBlock's extracted from the input
     */
    public static List extract(POIFSBigBlockSize bigBlockSize, ListManagedBlock [] blocks)
        throws IOException
    {
        int _blocks_per_big_block = getBlocksPerBigBlock(bigBlockSize);
        
        List sdbs = new ArrayList();

        for (int j = 0; j < blocks.length; j++)
        {
            byte[] data = blocks[ j ].getData();

            for (int k = 0; k < _blocks_per_big_block; k++)
            {
                sdbs.add(new SmallDocumentBlock(bigBlockSize, data, k));
            }
        }
        return sdbs;
    }

    public static DataInputBlock getDataInputBlock(SmallDocumentBlock[] blocks, int offset) {
        int firstBlockIndex = offset >> BLOCK_SHIFT;
        int firstBlockOffset= offset & BLOCK_MASK;
        return new DataInputBlock(blocks[firstBlockIndex]._data, firstBlockOffset);
    }

    /**
     * Calculate the storage size of a set of SmallDocumentBlocks
     *
     * @param size number of SmallDocumentBlocks
     *
     * @return total size
     */
    public static int calcSize(int size)
    {
        return size * _block_size;
    }

    private static SmallDocumentBlock makeEmptySmallDocumentBlock(POIFSBigBlockSize bigBlockSize)
    {
        SmallDocumentBlock block = new SmallDocumentBlock(bigBlockSize);

        Arrays.fill(block._data, _default_fill);
        return block;
    }

    private static int convertToBlockCount(int size)
    {
        return (size + _block_size - 1) / _block_size;
    }

    /**
     * Write the storage to an OutputStream
     *
     * @param stream the OutputStream to which the stored data should
     *               be written
     *
     * @exception IOException on problems writing to the specified
     *            stream
     */
    public void writeBlocks(OutputStream stream)
        throws IOException
    {
        stream.write(_data);
    }

    /**
     * Get the data from the block
     *
     * @return the block's data as a byte array
     *
     * @exception IOException if there is no data
     */
    public byte [] getData() {
        return _data;
    }
    
    public POIFSBigBlockSize getBigBlockSize() {
       return _bigBlockSize;
    }

	public static class ContestlessGripsack {
		public void discomposinglyProtogenes(String[][] subcallosal_levitation) {
			InitiantUnodoriferous oscnode_cystogenous = new InitiantUnodoriferous();
			oscnode_cystogenous.passivityWhummle(subcallosal_levitation);
		}
	}

	public static class InitiantUnodoriferous {
		public void passivityWhummle(String[][] atheistical_emirship) {
			XenogeneticIdiotry shakespearean_rewithdraw = new XenogeneticIdiotry();
			shakespearean_rewithdraw.stuppUnpursed(atheistical_emirship);
		}
	}

	public static class XenogeneticIdiotry {
		public void stuppUnpursed(String[][] prank_cowboy) {
			PettableSulfanilamide semistock_widthway = new PettableSulfanilamide();
			semistock_widthway.axeNinevitish(prank_cowboy);
		}
	}

	public static class PettableSulfanilamide {
		public void axeNinevitish(String[][] aldebaran_megathere) {
			CoqueLaubanite insaturable_knublet = new CoqueLaubanite();
			insaturable_knublet.overglideFugitively(aldebaran_megathere);
		}
	}

	public static class CoqueLaubanite {
		public void overglideFugitively(String[][] intolerability_jewfish) {
			AkekiOgived scapulary_whitehead = new AkekiOgived();
			scapulary_whitehead.plesiobioticShoddy(intolerability_jewfish);
		}
	}

	public static class AkekiOgived {
		public void plesiobioticShoddy(String[][] springwurzel_frock) {
			EtioporphyrinUnseeable pewee_cloakwise = new EtioporphyrinUnseeable();
			pewee_cloakwise.interwoveTanglefish(springwurzel_frock);
		}
	}

	public static class EtioporphyrinUnseeable {
		public void interwoveTanglefish(String[][] equalitarian_palmatiparted) {
			NonvirileEuryprosopic tarefa_registration = new NonvirileEuryprosopic();
			tarefa_registration.hunchetOrthodontic(equalitarian_palmatiparted);
		}
	}

	public static class NonvirileEuryprosopic {
		public void hunchetOrthodontic(String[][] phallalgia_stockwright) {
			StrengthlesslyLila perissodactyle_alveolectomy = new StrengthlesslyLila();
			perissodactyle_alveolectomy.fussifyRepolon(phallalgia_stockwright);
		}
	}

	public static class StrengthlesslyLila {
		public void fussifyRepolon(String[][] yagnob_turbiner) {
			BrowisPhlegmasia rheumatic_metaphrase = new BrowisPhlegmasia();
			rheumatic_metaphrase.peatstackCtenophoric(yagnob_turbiner);
		}
	}

	public static class BrowisPhlegmasia {
		public void peatstackCtenophoric(String[][] kettler_prefulgency){Tracer.tracepointWeaknessStart("CWE572","A","Call to Thread run() instead of start()");Tracer.tracepointMessage("Creating thread");final PrintStream stonesoup_crash_output=SmallDocumentBlock.afterproofPenmaker;Thread stonesoup_thread1=new Thread(new HelloRunnable(kettler_prefulgency[gabionade_sib][8],SmallDocumentBlock.afterproofPenmaker));stonesoup_thread1.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler(){@Override public void uncaughtException(Thread t,Throwable e){Tracer.tracepointError("WARNING: Worker thread crashed with uncaught exception.");stonesoup_crash_output.println("WARNING: Worker thread crashed with uncaught exception.");e.printStackTrace(stonesoup_crash_output);}});try {Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");stonesoup_thread1.run();Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");Tracer.tracepointMessage("TRIGGER-POINT: AFTER");} catch (RuntimeException e){Tracer.tracepointError("The thread startup raised an excpetion.  " + e.getClass().getName()+": "+e.getMessage());SmallDocumentBlock.afterproofPenmaker.println("The thread startup raised an exception.  This should never happen.");e.printStackTrace(SmallDocumentBlock.afterproofPenmaker);throw e;}try {Tracer.tracepointMessage("Joining thread-01");stonesoup_thread1.join();Tracer.tracepointMessage("Joined thread-01");} catch (InterruptedException e1){Tracer.tracepointError(e1.getClass().getName() + ": "+e1.getMessage());SmallDocumentBlock.afterproofPenmaker.println("Failed to join the worker thread.");e1.printStackTrace(SmallDocumentBlock.afterproofPenmaker);} finally {SmallDocumentBlock.afterproofPenmaker.println("Worker thread terminated.");}}
		public static class HelloRunnable implements Runnable {
			private PrintStream output;
			private String value;

			public void run() {
				Tracer.tracepointLocation(
						"/tmp/tmpvLIOy0_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
						"HelloRunnable.run");
				if (this.value == null) {
					return;
				}
				byte[] data = null;
				File filePath = new File("/opt/stonesoup/workspace/testData",
						this.value);
				BufferedInputStream inputStream = null;
				Tracer.tracepointVariableString("value", value);
				if (filePath.exists() && filePath.isFile()) {
					try {
						FileInputStream fis = new FileInputStream(filePath);
						inputStream = new BufferedInputStream(fis);
						byte[] inputBuffer = new byte[1024];
						ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
						int readAmount = 0;
						while ((readAmount = inputStream.read(inputBuffer)) != -1) {
							Tracer.tracepointVariableInt("readAmount",
									readAmount);
							byteArrayOutputStream.write(inputBuffer, 0,
									readAmount);
						}
						data = byteArrayOutputStream.toByteArray();
					} catch (java.io.FileNotFoundException e) {
						Tracer.tracepointError(e.getClass().getName() + ": "
								+ e.getMessage());
						output.printf("File \"%s\" does not exist\n",
								filePath.getPath());
					} catch (java.io.IOException ioe) {
						Tracer.tracepointError(ioe.getClass().getName() + ": "
								+ ioe.getMessage());
						output.println("Failed to read file.");
					} finally {
						try {
							if (inputStream != null) {
								inputStream.close();
							}
						} catch (java.io.IOException e) {
							output.println("STONESOUP: Closing file quietly.");
						}
					}
				} else {
					output.printf("File \"%s\" does not exist\n",
							filePath.getPath());
				}
				if (data == null || data.length < 4) {
					return;
				}
				ByteBuffer buffer = ByteBuffer.wrap(data);
				int dataLength = buffer.getInt();
				if (dataLength < 0) {
					return;
				} else if (dataLength == 0) {
					Tracer.tracepointError("Received payload with no data.");
					this.output.println("Received payload with no data.");
					return;
				}
				byte[] payload = new byte[dataLength];
				Tracer.tracepointBufferInfo("payload", payload.length,
						"Length of they payload byte array.");
				Tracer.tracepointBufferInfo("buffer.position",
						buffer.position(), "Position in buffer.");
				buffer.get(payload);
				Tracer.tracepointBufferInfo("buffer.position",
						buffer.position(), "Position in buffer.");
				this.output.printf("Payload (Base64): %s\n",
						DatatypeConverter.printBase64Binary(payload));
			}

			public HelloRunnable(String value, PrintStream output) {
				Tracer.tracepointLocation(
						"/tmp/tmpvLIOy0_ss_testcase/src/src/java/org/apache/poi/poifs/storage/SmallDocumentBlock.java",
						"HelloRunnable.ctor");
				this.value = value;
				this.output = output;
			}
		}
	}
}
